# -*- coding: latin1 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
# 
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS - 
# Web-Anwendungs-Server fuer Kompetenzagenturen).
# 
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
# 
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
# 
# mpuls WASKA has been developed on behalf of the 
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and 
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

from nodecomponents import \
    NodeComponent, \
    Node, \
    Leaf, \
    VISIT_IGNORE_CHILDREN, \
    VISIT_CONTINUE

from exprtree import Expr

import os
import sys

from datetime import date

class FilteredNodeComponentProxy:
    def __init__(self, params):
        self.params = params

    def set(self, nc):
        self.nc = nc

    def getName(self):
        return self.nc.getName()

    def allWidgets(self):
        return [w for w in self.nc.allWidgets() if w.getName() in self.params]

class ConditionalWidgetCollector(object):

    def __init__(self, evaluator = lambda expr: True):
        self.widgets   = []
        self.evaluator = evaluator

    def visitor(self, nc):
        if isinstance(nc, Node):
            if isinstance(nc, ConditionalNode):
                expr = nc.getExpr()
                if expr and not self.evaluator(expr):
                    return VISIT_IGNORE_CHILDREN
            elif isinstance(nc, (RadioNode, ChoiceNode)):
                self.widgets.append(nc)
                return VISIT_IGNORE_CHILDREN
            elif isinstance(nc, GroupNode) and nc.isRepeat() and nc != self:
                return VISIT_IGNORE_CHILDREN
        elif isinstance(nc, Leaf) and not isinstance(nc,
            (RuleLeaf, ExternalChoiceListLeaf, InfoLeaf)):
            self.widgets.append(nc)

        return VISIT_CONTINUE
        
class WidgetCollector(object):

    def __init__(self):
        self.widgets = []

    def visitor(self, nc):
        if isinstance(nc, Node):
            if isinstance(nc, (RadioNode, ChoiceNode)):
                self.widgets.append(nc)
                return VISIT_IGNORE_CHILDREN
            if isinstance(nc, GroupNode) and nc.isRepeat() and nc != self:
                return VISIT_IGNORE_CHILDREN
        elif isinstance(nc, Leaf) and not isinstance(nc, 
            (RuleLeaf, ExternalChoiceListLeaf, InfoLeaf)):
            self.widgets.append(nc)

        return VISIT_CONTINUE


class PageNode(Node):
    def __init__(self):
        Node.__init__(self)

    def allWidgets(self):
        collector = WidgetCollector()
        self.visit(collector.visitor)
        return collector.widgets

class GroupNode(Node):
    def __init__(self):
        Node.__init__(self)
        self.attributes["repeat"]  = ""
        self.attributes["invisible"] = ""

    def isRepeat(self):
        return self.getAttribute("repeat") and True or False

    def setRepeat(self, repeat):
        repeat = repeat and "true" or ""
        self.setAttribute("repeat", repeat)

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False

    def setInvisible(self, invisible):
        invisible = invisible and "true" or ""
        self.setAttribute("invisible", invisible)

    def allWidgets(self):
        collector = WidgetCollector()
        self.visit(collector.visitor)
        return collector.widgets

class ConditionalNode(Node):

    def __init__(self):
        Node.__init__(self)
        self.attributes["expr"]      = ""
        self.attributes["invisible"] = ""
        self.expr = None

    def setAttribute(self, attrib, value, broadcast=True):
        if attrib == "expr":
            newExpr = Expr(value)
            newExpr.compile()
            self.expr = newExpr
        Node.setAttribute(self, attrib, value, broadcast)

    def setExpr(self, expr):
        self.setAttribute('expr', expr)

    def getExpr(self):
        return self.expr

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False

    def setInvisible(self):
        invisible = invisible and "true" or ""
        self.setAttribute("invisible", invisible)

class MatrixNode(GroupNode):
    def __init__(self):
        GroupNode.__init__(self)

class RadioNode(Node):
    def __init__(self):
        Node.__init__(self)
        self.attributes["selected"] = ""

    def firstFindDefault(self):
        value = None
        if self.children:
            for child in self.children:
                if isinstance(child, ExternalChoiceListLeaf):
                    found, v = child.find_default()
                    if found: 
                        value = v
                        break
                elif isinstance(child, BoolLeaf):
                    if child.getChecked():
                        v = child.getValue()
                        if v: value = int(v)
                        break

        self.found_default = value
        return Node.firstFindDefault(self)

    def getSelected(self):
        return self.getAttribute("selected")

    def setSelected(self, selected):
        self.setSelected("selected", selected)

class BoolLeaf(Leaf):
    def __init__(self):
        Leaf.__init__(self)
        self.attributes["checked"] = ""

    def getChecked(self):
        return self.getAttribute("checked")

    def setChecked(self, checked):
        self.setSelected("checked", checked)
    
class ExternalChoiceListLeaf(Leaf):
    def __init__(self):
        Leaf.__init__(self)
        self.choice = None
        self.path   = None

    def setLoadContext(self, path):
        self.path = os.path.dirname(path)

    def find_default(self):
        found, value = False, None
        children = self.getChildren()
        if children:
            for child in children:
                if isinstance(child, BoolLeaf):
                    if child.getChecked():
                        v = child.getValue()
                        if v: value = int(v)
                        break
        return found, value

    def getChildren(self):
        if not self.choice:
            fname = self.getValue()
            if fname:
                if not os.path.isabs(fname) and self.path:
                    fname = os.path.join(self.path, fname)
                try: # TODO: This loading code should be else where!
                    from offline.io.document import openDocument
                    document = openDocument(fname)
                    self.choice = document.findByClassAndName(ChoiceNode)
                except Exception, inst:
                    print str(inst)
        if self.choice: 
            return self.choice.children
        return None

class InfoLeaf(Leaf):
    def __init__(self):
        Leaf.__init__(self)
        self.attributes["invisible"] = ""

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False

    def setInvisible(self, invisible):
        invisible = invisible and "true" or ""
        self.setAttribute("invisible", invisible)

class TextAreaLeaf(Leaf):
    def __init__(self):
        Leaf.__init__(self)
        self.attributes["cols"] = ""
        self.attributes["rows"] = ""

    def firstFindDefault(self):
        value = self.getValue()
        if not value: value = None
        self.found_default = value
        return Leaf.firstFindDefault(self)

    def getCols(self):
        return self.getAttribute("cols")

    def setCols(self, cols):
        return self.setAttribute("cols", cols)

    def getRows(self):
        return self.getAttribute("rows")

    def setRows(self, rows):
        return self.setAttribute("rows", rows)

class TextLeaf(Leaf):
    def __init__(self):
        Leaf.__init__(self)
        self.attributes["size"] = "40"
        self.attributes["maxlength"] = "60"

    def firstFindDefault(self):
        value = self.getValue()
        if not value: value = None
        self.found_default = value
        return Leaf.firstFindDefault(self)

    def getSize(self):
        return self.getAttribute("size")

    def setSize(self, size):
        self.setAttribute("size", size)

    def getMaxLength(self):
        return self.getAttribute("maxlength")

    def setMaxLength(self, length):
        self.setAttribute("maxlength", length)

class ChoiceNode(Node):
    def __init__(self):
        Node.__init__(self)
        self.attributes["size"]     = "1"
        self.attributes["multiple"] = ""

    def firstFindDefault(self):
        value = None
        if self.children:
            for child in self.children:
                if isinstance(child, ExternalChoiceListLeaf):
                    found, v = child.find_default()
                    if found: 
                        value = v
                        break
                elif isinstance(child, BoolLeaf):
                    if child.getChecked():
                        v = child.getValue()
                        if v: value = int(v)
                        break

        self.found_default = value
        return Node.firstFindDefault(self)

    def getSize(self):
        return self.getAttribute("size")

    def setSize(self, size):
        self.setAttribute("size", size)

    def getMultiple(self):
        return self.getAttribute("multiple")

    def setMultiple(self, multiple):
        self.setAttribute("multiple", multiple)

    def getSelected(self):
        return self.getAttribute("selected")

    def setSelected(self, selected):
        self.setAttribute("selected", selected)

class IntLeaf(Leaf):
    def __init__(self):
        Leaf.__init__(self)
        self.attributes["minvalue"] = ""
        self.attributes["maxvalue"] = ""

    def firstFindDefault(self):
        value = self.getValue().strip()
        if value:
            try:
                value = int(value)
            except ValueError:
                print >> sys.stderr, "Broken default for '%s'" % self.getName()
                value = None
        else:
            value = None
        self.found_default = value
        return Leaf.firstFindDefault(self)

    def getMinValue(self):
        return self.getAttribute("minvalue")

    def setMinValue(self, minValue):
        self.setAttribute("minvalue", minValue)

    def getMaxValue(self):
        return self.getAttribute("maxvalue")

    def setMaxValue(self, maxValue):
        self.setAttribute("maxvalue", maxValue)

class DateLeaf(Leaf):

    def __init__(self):
        Leaf.__init__(self)

    def firstFindDefault(self):
        value = self.getValue().strip()
        if value:
            try:
                value = date(*[int(x) for x in value.split("-")])
            except (TypeError, ValueError):
                print >> sys.stderr, "Broken default for '%s'" % self.getName()
                value = None
        else:
            value = None
        self.found_default = value
        return Leaf.firstFindDefault(self)

class RuleLeaf(Leaf):
    def __init__(self):
        Leaf.__init__(self)
        self.attributes["expr"] = ""
        self.attributes["mark"] = ""
        self.expr = None

    def shallowCopy(self):
        r = Leaf.shallowCopy(self)
        if self.expr: r.expr = self.expr.clone()
        else:         r.expr = None
        return r

    def setAttribute(self, attrib, value, broadcast=True):
        if attrib == "expr":
            newExpr = Expr(value)
            newExpr.compile()
            self.expr = newExpr
        Leaf.setAttribute(self, attrib, value, broadcast)

    def getExpr(self):
        return self.expr

    def setExpr(self, expr):
        self.setAttribute("expr", expr)

    def getMark(self):
        return self.getAttribute("mark")

    def setMark(self, mark):
        self.setAttribute("mark", mark)

# vim:set ts=4 sw=4 si et sta sts=4:
