# -*- coding: utf-8 -*-
#
# (c) 2008 by Intevation GmbH
# This is Free software under the GPLv3. 
# See LICENSE comming with the source of 'mpuls offline'
# for details.
#
# author: Sascha L. Teichmann <sascha.teichmann@intevation.de>
#

from xml.sax          import make_parser
from xml.sax.saxutils import escape

from uuid import uuid4

import datetime
import sys

from offline.lib.xmlhelper import \
    StackedLoader,                \
    decode_int,                   \
    iso_datetime_string,          \
    decode_iso_datetime

from offline.lib.helpers import \
    dd_mm_YYYY,                 \
    HH_MM,                      \
    parse_HH_MM,                \
    parse_dd_mm_YYYY,           \
    parse_pos_int,              \
    safe_unicode


class EntryType(object):

    def __init__(self, description, value, is_editable):
        self.description = description
        self.value       = value
        self.is_editable = is_editable

    def getDescription(self):
        return self.description

    def isEditable(self):
        return self.is_editable

    def getValue(self):
        return self.value

class LogbookMeta(object):

    def __init__(self, filename = None):
        self.entry_types = []
        if filename: self.load(filename)

    def default_start(self, name, attrs):
        pass

    def default_end(self, name, attrs, value):
        pass

    def entry_type_end(self, name, attrs, text):
        value = attrs.get(u'value')
        if value is None: return
        editable = attrs.get('editable', 'true').lower() != 'false'
        self.entry_types.append(
            EntryType(text.strip(), int(value), editable))

    def load(self, filename):
        f = None
        try:
            loader = StackedLoader()

            factories = {
                u'entry-type': (self.default_start, self.entry_type_end) 
            }

            loader.add_factories(
                u'logbook-meta',
                factories,
                (self.default_start, self.default_end))

            parser = make_parser()
            parser.setContentHandler(loader)

            f = open(filename.encode(sys.getfilesystemencoding()), "rb")
            parser.parse(f)
        finally:
            if f:
                try: f.close()
                except: pass

    def find_entry_type_by_value(self, value):
        for entry_type in self.entry_types:
            if entry_type.value == value:
                return entry_type
        return None

class LogbookEntryErrors(object):

    def __init__(self, params, meta = None):

        entry = LogbookEntry()
        self.entry  = entry
        self.errors = []

        def param(k):
            v = params.get(k, ('',))[0]
            # XXX: Python and the Holy Encoding Grail
            for enc in ('utf-8', 'latin-1'):
                try:
                    return unicode(v, enc).strip()
                except UnicodeDecodeError:
                    pass
            return safe_unicode(v).strip()

        time = param('time')

        h, m = 0, 0
        if time:
            try:
                h, m = parse_HH_MM(time)
            except ValueError, e:
                self.errors.append(('time', e.message))

        date = param('date')

        if not date:
            date = datetime.date.today().strftime("%d.%m.%Y")

        try:
            d, mon, y = parse_dd_mm_YYYY(date)
            try:
                entry.datum = datetime.datetime(y, mon, d, h, m)
            except ValueError:
                raise ValueError(u'Kein gültiges Datum')
        except ValueError, e:
            self.errors.append(('date', e.message))

        duration = param('duration')

        if duration:
            try:
                entry.setDuration(parse_pos_int(duration))
            except ValueError, e:
                self.errors.append(('duration', e.message))

        editor = param('editor')
        if editor: entry.setEditor(editor)

        short_notice = param('short_notice')
        if short_notice: entry.setShortNotice(short_notice)

        notice = param('notice')
        if notice: entry.setNotice(notice)

        uuid = param('uuid')
        if uuid: entry.setUUID(uuid)

        if meta:
            kind = decode_int(param('kind'))
            if not kind is None:
                kind = meta.find_entry_type_by_value(kind)
                if kind: entry.setType(kind)

    def getEntry(self):
        return self.entry

    def hasErrors(self):
        return bool(self.errors)

    def hasError(self, field):
        for error in self.errors:
            if error[0] == field:
                return True
        return False

    def getErrors(self):
        return self.errors

def _cmp(a, b):
    if a is None and b is None: return 0
    if a is None: return -1
    if b is None: return +1
    return cmp(a, b)

def ORDER_ENTRIES_BY_DATE_ASC(a, b):
    return _cmp(a.datum, b.datum)

def ORDER_ENTRIES_BY_DATE_DESC(a, b):
    return _cmp(b.datum, a.datum)

class LogbookEntry(object):
    
    def __init__(self, uuid = None):
        self.uuid       = uuid
        self.datum      = None
        self.art        = None
        self.bearbeiter = None
        self.kurz_notiz = None
        self.notiz      = None
        self.dauer      = None

    def getUUID(self):
        if not self.uuid:
            self.uuid = str(uuid4())
        return self.uuid

    def setUUID(self, uuid):
        self.uuid = uuid

    def setDate(self, datum):
        self.datum = datum

    def getDate(self):
        return dd_mm_YYYY(self.datum)

    def getTime(self):
        return HH_MM(self.datum)

    def setEditor(self, editor):
        self.bearbeiter = editor

    def getEditor(self):
        return u"" if self.bearbeiter is None else self.bearbeiter

    def getType(self):
        return self.art

    def setType(self, art):
        self.art = art

    def getDuration(self):
        return u"" if self.dauer is None else "%d" % self.dauer

    def setDuration(self, duration):
        self.dauer = duration

    def getShortNotice(self):
        return u"" if self.kurz_notiz is None else self.kurz_notiz

    def setShortNotice(self, short_notice):
        self.kurz_notiz = short_notice

    def getNotice(self):
        return u"" if self.notiz is None else self.notiz

    def setNotice(self, notice):
        self.notiz = notice

    def to_xml(self, out):
        out.append(u'<tagebuch-eintrag>')

        if not self.uuid is None:
            out.append(u'<uuid_id>%s</uuid_id>' % self.uuid)

        if not self.datum is None:
            out.append(u'<tagebuch-datum>%s</tagebuch-datum>' % \
                iso_datetime_string(self.datum))

        if not self.bearbeiter is None:
            out.append(u'<tagebuch-bearbeiter>%s</tagebuch-bearbeiter>' % \
                escape(self.bearbeiter))

        if not self.art is None:
            out.append(u'<tagebuch-art>%d</tagebuch-art>' % \
                self.art.value)

        if not self.kurz_notiz is None:
            out.append(u'<tagebuch-sonstiges>%s</tagebuch-sonstiges>' % \
                escape(self.kurz_notiz))

        if not self.notiz is None:
            out.append(u'<tagebuch-notiz>%s</tagebuch-notiz>' % \
                escape(self.notiz))

        if not self.dauer is None:
            out.append(u'<tagebuch-dauer>%d</tagebuch-dauer>' % self.dauer)

        out.append(u'</tagebuch-eintrag>')

class Logbook(object):

    def __init__(self):
        self.entries = []

    def to_xml(self):
        if not self.entries: return u""
        out = [u"<tagebuch>"]
        for entry in self.entries:
            entry.to_xml(out)
        out.append(u"</tagebuch>")
        return u"".join(out)

    def sorted_entries(self, cmp = ORDER_ENTRIES_BY_DATE_ASC):
        entries = self.entries[:]
        entries.sort(cmp)
        return entries

    def add_entry(self, entry):
        self.entries.append(entry)

    def delete_entry_by_uuid(self, uuid):
        for idx, entry in enumerate(self.entries):
            if entry.uuid == uuid:
                self.entries.pop(idx)
                return True
        return False

    def replace_entry_by_uuid(self, new_entry):
        nuuid = new_entry.uuid
        for idx, entry in enumerate(self.entries):
            if entry.uuid == nuuid:
                self.entries[idx] = new_entry
                return True
        return False

    def get_entry_by_uuid(self, uuid):
        for entry in self.entries:
            if entry.uuid == uuid:
                return entry
        return None

class LogbookLoader(object):

    def __init__(self, meta):
        self.logbook = Logbook()
        self.meta    = meta
        self.entry   = None

    def register_loader(self, loader):

        factories = {
            u"tagebuch":            (self.default_start, self.default_end),
            u"tagebuch-eintrag":    (self.eintrag_start, self.eintrag_end),
            u"uuid_id":             (self.default_start, self.uuid_end),
            u'tagebuch-datum':      (self.default_start, self.datum_end),
            u'tagebuch-art':        (self.default_start, self.art_end),
            u'tagebuch-sonstiges':  (self.default_start, self.kurz_end),
            u'tagebuch-bearbeiter': (self.default_start, self.bearbeiter_end),
            u'tagebuch-notiz':      (self.default_start, self.notiz_end),
            u'tagebuch-dauer':      (self.default_start, self.dauer_end),
        }

        loader.add_factories(
            u"tagebuch", 
            factories,
            (self.default_start, self.default_end))

    def art_end(self, name, attrs, value):
        value = decode_int(value)
        if not value is None:
            entry_type = self.meta.find_entry_type_by_value(value)
            if self.entry and entry_type:
                self.entry.setType(entry_type)

    def dauer_end(self, name, attrs, value):
        if self.entry: self.entry.dauer = decode_int(value)

    def bearbeiter_end(self, name, attrs, value):
        if self.entry: self.entry.bearbeiter = value

    def notiz_end(self, name, attrs, value):
        if self.entry: self.entry.notiz = value

    def kurz_end(self, name, attrs, value):
        if self.entry: self.entry.kurz_notiz = value

    def datum_end(self, name, attrs, value):
        if self.entry:
            self.entry.datum = decode_iso_datetime(value.strip())

    def eintrag_start(self, name, attrs):
        self.entry = LogbookEntry()

    def eintrag_end(self, name, attrs, value):
        if not self.entry.uuid:
            self.entry.uuid = str(uuid4())
        self.logbook.entries.append(self.entry)
        self.entry = None

    def uuid_end(self, name, attrs, value):
        if self.entry: self.entry.uuid = value.strip()

    def default_start(self, name, attrs):
        pass

    def default_end(self, name, attrs, value):
        pass

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
