# -*- coding: utf-8 -*-
#
# (c) 2008 by Intevation GmbH
# This is Free software under the GPLv3. 
# See LICENSE comming with the source of 'mpuls offline'
# for details.
#
# author: Sascha L. Teichmann <sascha.teichmann@intevation.de>
#

from xml.sax          import make_parser
from xml.sax.saxutils import escape

from datetime         import date
from traceback        import print_exc

import codecs
import sys
import data 
import re
import shutil

from uuid import uuid4

from semantic import SematicError, checkAndConvert, ErrorItem

from data import PageNode, RuleLeaf, WidgetCollector, ConditionalWidgetCollector

from navigation import Navigation

from logbook import LogbookLoader, Logbook

from offline.lib.xmlhelper import StackedLoader, decode_int, decode_date

import os.path

# version of the loaded XML file
VERSION = 3

WARNING    = re.compile(r"warning\s*:\s*([^\s]+)")

DECODERS = {
    data.RadioNode:    decode_int,
    data.TextAreaLeaf: lambda s: s,
    data.TextLeaf:     lambda s: s,
    data.ChoiceNode:   decode_int,
    data.IntLeaf:      decode_int,
    data.DateLeaf:     decode_date
}

ENCODERS = {
    date     : lambda x: "%d-%d-%d" % (x.year, x.month, x.day),
    str      : lambda x: escape(x),
    type(u''): lambda x: escape(x)
}

XML_HEADER = \
u'''<?xml version="1.0" encoding="UTF-8"?>
<cases version="%s">
<case>'''

XML_FOOTER = \
u'''</case>
</cases>'''

class MasterLoader(object):

    def __init__(self, widgets, converter = None):
        self.widgets   = widgets
        self.new_data  = {}
        self.converter = converter
        self.version   = 1
        self.errors    = {}

    def element_start(self, name, attrs):
        if name == u"cases":
            try:
                self.version = int(attrs[u'version'])
                if self.version > VERSION:
                    raise Exception(
                        "Cannot load files of version %d (only up to %d)." % (
                            self.version, VERSION))
            except (KeyError, ValueError):
                pass

    def element_end(self, name, attrs, value):

        if self.converter and self.converter.eat(name, attrs, value):
            return

        try:
            widget = self.widgets[name]
        except KeyError:
            if name == u'uuid_id':
                self.new_data[name] = value
            #print >> sys.stderr, "Tag not found: %s" % repr(name)
            return

        if value == u'':
            self.new_data[name] = None
        else:
            try:
                value = checkAndConvert(widget, value)
                self.new_data[name] = value
            except SematicError, inst:

                try:
                    ei = self.errors[name]
                except KeyError:
                    page = widget.getParentOfClass(data.PageNode)
                    ei = ErrorItem(
                        page.getName(),
                        value, 
                        widget.getDescription())
                    self.errors[name] = ei
                ei.addMessage(inst.value)

class InconsistencyItem(object):

    def __init__(self, rule, navigation, widgets):
        self.rule = rule
        links = []
        for dep in rule.getExpr().getDependencies():
            try:
                widget = widgets[dep]
            except KeyError:
                continue
            page = widget.parent
            while not page is None and not isinstance(page, PageNode):
                page = page.parent
            if not page: continue
            item = navigation.getTreeItem(page.getName())
            if not item: continue
            select_link = item.get_select_link("#f_%s" % dep)
            link = u'<a href="%s">%s</a>' % (
                select_link, escape(widget.getDescription()))
            links.append(link)
        self.links = u", ".join(links)

class MemoryDatastore(object):

    def __init__(
        self, 
        formed, 
        defaults, 
        logbook_meta, 
        converter_factory = None,
        filename          = None
    ):
        self.formed            = formed
        self.data              = {}
        self.dirty             = False
        self.defaults          = defaults
        self.logbook_meta      = logbook_meta
        self.converter_factory = converter_factory
        self.navigation        = None

        collector = WidgetCollector()
        formed.visit(collector.visitor)
        self.widgets = dict([(w.getName(), w) for w in collector.widgets if w.getName()])

        if filename:
            self.load(filename)
            self.inconsistencies = self.check_inconsistencies()
        else:
            self.inconsistencies = []
            self.errors          = {}
            self.logbook = Logbook()
            self.presetDefaults()

        self.filename = filename
        self.swapfile = None

    def hasData(self, name):
        return self.data.has_key(name)

    def setConverterFactory(self, converter_factory):
        self.converter_factory = converter_factory

    def setSwapfile(self, swapfile=None):
        self.swapfile = swapfile

    def presetDefaults(self):
        defaults = self.defaults
        for item in defaults.values:
            try:
                w = self.widgets[item.key]
            except KeyError:
                continue
            try:
                iv = item.eval()
            except:
                print_exc()
                continue
            decoder = DECODERS.get(w.__class__, lambda s: s)
            try:
                iv = decoder(iv)
                if iv and w.findDefault() != iv:
                    self.data[item.key] = iv
            except:
                print_exc()

    def isDirty(self):
        return self.dirty

    def setDirty(self, flag):
        self.dirty = flag

    def getData(self, name):
        try:
            return self.data[name]
        except KeyError:
            try:
                value = self.widgets[name].findDefault()
                self.data[name] = value
                return value
            except KeyError:
                return None

    def getFilename(self):
        return self.filename

    def getErrors(self):
        return self.errors

    def makeUnique(self):
        try:
            uuid_id = self.data['uuid_id']
            nuuid_id = str(uuid4())
            while nuuid_id == uuid_id:
                nuuid_id = str(uuid4())
            self.data['uuid_id'] = nuuid_id
        except KeyError:
            self.data['uuid_id'] = str(uuid4())

        self.dirty = True


    def getPageNamesWithErrors(self):
        return set([e.page for e in self.errors.itervalues()])

    def getNavigation(self):
        if self.navigation is None:
            self.navigation = Navigation(self.formed)
        return self.navigation

    def deleteData(self, name):
        try:
            v = self.data[name]
            d = self.widgets[name].findDefault()
            if v != d:
                self.data[name] = d
                self.dirty = True
                return True
            return False
        except KeyError:
            return False

    def setData(self, name, value):
        try:
            old = self.data[name]
            if old != value:
                self.data[name] = value
                self.dirty = True
                return True
            return False
        except KeyError:
            w = self.widgets.get(name)
            if w and w.findDefault() != value:
                self.data[name] = value
                self.dirty = True
                return True
        return False

    def setFilename(self, filename):
        if self.filename != filename:
            self.filename = filename
            self.dirty = True

    def getLogbook(self):
        if not self.logbook:
            self.logbook = Logbook()
        return self.logbook

    def deleteLogbookEntryByUUID(self, uuid):
        logbook = self.getLogbook()
        if logbook.delete_entry_by_uuid(uuid):
            self.dirty = True
            return True
        return False

    def replaceLogbookEntryByUUID(self, entry):
        logbook = self.getLogbook()
        if logbook.replace_entry_by_uuid(entry):
            self.dirty = True
            return True
        return False

    def addEntryToLogbook(self, entry):
        logbook = self.getLogbook()
        logbook.add_entry(entry)
        self.dirty = True

    def load(self, filename):

        converter = self.converter_factory and self.converter_factory() or None

        master_loader = MasterLoader(self.widgets, converter)

        f = None
        try:
            f = open(filename.encode(sys.getfilesystemencoding()), "rb")
            parser = make_parser()
            loader = StackedLoader()
            loader.add_factories(
                'master',
                {}, 
                (master_loader.element_start, master_loader.element_end))
            loader.activate_factories('master') # for backwards compat

            logbook_loader = LogbookLoader(self.logbook_meta)
            logbook_loader.register_loader(loader)

            parser.setContentHandler(loader)
            parser.parse(f)

            if converter: converter.finish(master_loader)

            new_data = master_loader.new_data

            self.logbook = logbook_loader.logbook
        finally:
            if f:
                try: f.close()
                except: pass


        self.filename = filename
        self.data     = new_data
        self.dirty    = False
        self.errors   = master_loader.errors

    def getInconsistencies(self):
        return self.inconsistencies

    def update_inconsistencies(self):
        if not self.inconsistencies:
            return
        new_inconsistencies = []
        D = self.getData
        for inconsistency in self.inconsistencies:
            expr = inconsistency.rule.getExpr()
            params, isNull = {}, False
            for dep in expr.getDependencies():
                v = D(dep)
                if v is None: isNull = True; break
                params[dep] = v
            if isNull: continue # only eval rule if all vars are not NULL
            try:
                okay = expr.evaluate(params)
            except:
                okay = False
            if not okay:
                new_inconsistencies.append(inconsistency)
        self.inconsistencies = new_inconsistencies

    def check_inconsistencies(self):
        ''' check all error rules'''
        errors = []
        affected_vars = set()
        D = self.getData
        for r in self.formed.findAllByClass(RuleLeaf):
            mark, expr = r.getMark(), r.getExpr()
            # warning rules are evaluated at page rendering
            if mark and (mark.find("warning:") >= 0 or mark.find("notice:") >= 0):
                continue
            params, isNull = {}, False
            dependencies = expr.getDependencies()
            for dep in expr.getDependencies():
                v = D(dep)
                if v is None: isNull = True; break
                params[dep] = v
            if isNull: continue # only eval rule if all vars are not NULL
            try:
                okay = expr.evaluate(params)
            except:
                okay = False
            if not okay:
                errors.append(r)
                affected_vars |= dependencies

        if errors: # sort by first dep var
            collector = WidgetCollector()
            self.formed.visit(collector.visitor)
            widget_indices = dict(
                [(w.getName(), idx) for idx, w in enumerate(collector.widgets)
                    if w.getName() in affected_vars])

            min_index = {}
            for r in errors:
                deps = r.getExpr().getDependencies()
                min_d = None
                for dep in deps:
                    dep_idx = widget_indices.get(dep, -1)
                    if min_d is None or dep_idx < min_d:
                        min_d = dep_idx
                min_index[r.getName()] = min_d

            errors.sort(lambda a, b: \
                cmp(min_index[a.getName()], min_index[b.getName()]))

        navigation = self.getNavigation()
        errors = [InconsistencyItem(e, navigation, self.widgets) \
            for e in errors]
        return errors

    def store(self, filename=None, backup=True):
        if filename is None:
            if not self.filename: return
            filename = self.filename

        collector = WidgetCollector()
        self.formed.visit(collector.visitor)

        data = self.data

        # Write a backup file before storing data. The backup should always
        # contain the data of the last save
        if backup and os.path.isfile(filename):
            bak = filename + ".bak"
            try:
                shutil.copy(filename, bak)
            except OSError, e:
                pass

        out = None
        try:
            out = codecs.open(filename, "w", "UTF-8")
            print >> out, XML_HEADER % VERSION

            try:
                uuid_id = data['uuid_id']
            except KeyError:
                uuid_id = str(uuid4())
                data['uuid_id'] = uuid_id

            print >> out, u'<master>\n<uuid_id>%s</uuid_id>' % uuid_id

            for widget in collector.widgets:
                name = widget.getName()
                try:
                    value = data[name]
                    if value is None: continue
                except KeyError:
                    continue
                encoder = ENCODERS.get(type(value), lambda x: u"%s" % x)
                print >> out, u"<%s>%s</%s>" % (name, encoder(value), name)
            print >> out, u'</master>\n'

            if self.logbook:
                print >> out, self.logbook.to_xml()

            print >> out, XML_FOOTER

            if backup: self.dirty = False
        finally:
            if out:
                try: out.close()
                except: pass

    def evaluate(self, expr):
        D = self.getData
        try:
            return expr.evaluate(dict([(n, D(n)) for n in expr.getDependencies()]))
        except:
            return False

    def warningsOnPage(self, pageName):
        """ Generates a dictionary of warnings that should be displayed
            on a given page.
        """
        # find all widgets on page
        document = self.formed
        nc = document.findByClassAndName(PageNode, pageName)
        if not nc: return {}

        # widgets on current page
        used_fields = set([
            w.getName() for w in nc.allWidgets() if w.getName()])

        warnings = {}

        for r in document.findAllByClass(RuleLeaf):
            mark, expr = r.getMark(), r.getExpr()
            if not mark or not expr: continue
            # warnings only
            mark = WARNING.search(mark)
            if not mark: continue
            mark = mark.group(1)
            # only mark if field is on current page
            if not mark in used_fields: continue

            okay = self.evaluate(expr)

            if not okay:
                try:
                    t = warnings[mark]
                except KeyError:
                    t = (self.widgets[mark], [])
                    warnings[mark] = t
                t[1].append(r.getValue())

        return warnings

    def save(self, params, pageName):

        document = self.formed
        nc = document.findByClassAndName(PageNode, pageName)
        if not nc: return

        wc = ConditionalWidgetCollector(self.evaluate)
        nc.visit(wc.visitor)

        # widgets on current page
        widgets = dict([
            (widget.getName(), widget) for widget in wc.widgets if widget.getName()])

        widgets_clone = widgets.copy()

        dbPage = self
        old_errors = self.errors

        # empty parameters will nullify values
        delete_vars = []

        # vars that passed lexical and syntactic check
        to_be_set = {}

        for k, v in params.iteritems():
            if k.startswith("__"):
                continue
            try:
                value = v[0]
                widget = widgets.pop(k)
                #print "widget: %s"  % repr(widget)
                if value == "": # no value -> delete it
                    delete_vars.append(k)
                else:
                    # Hopefully this one causes no trouble any more
                    old_errors.pop(k, None)
                    # the empty array is for the rules
                    nv = checkAndConvert(widget, value)
                    ov = self.getData(k)
                    if nv != ov:
                        to_be_set[k] = (nv, [])
            except SematicError, inst:
                ei = ErrorItem(pageName, value, widget.getDescription())
                ei.addMessage(inst.value)
                # overwrite old errors
                old_errors[k] = ei
            except KeyError, inst:
                pass

        # flag to indicate if we should flush the page
        dirty = False

        # remove variable from dataset which are set to ""
        for wname in widgets.keys() + delete_vars:
            if dbPage.deleteData(wname):
                dirty = True
            # an unset value causes no problems.
            old_errors.pop(wname, None)

        # find all rules that depend on variables to be set.
        # build up a lookup var-name -> list of rules
        for r in document.findAllByClass(RuleLeaf):
            mark = r.getMark()
            # warning rules are evaluated at page rendering
            if mark and (mark.find("warning:") >= 0 or mark.find("notice:") >= 0):
                #print "Ignore warning rule '%s'" % r.getName()
                continue
            expr = r.getExpr()
            if not expr: continue
            for var in expr.getDependencies():
                try:
                    to_be_set[var][1].append(r)
                except KeyError:
                    pass

        # set vars into db page and check rules.
        for k, tup in to_be_set.iteritems():
            v, rules = tup[0], tup[1]
            hasNoProblems = True
            for rule in rules:
                vars = {}
                isNull = False # only process rule if all values are not null
                for dependency in rule.getExpr().getDependencies():
                    if dependency == k: # ignore because this comes from the web
                        continue
                    # XXX: Normally they are distributed over all kind of pages.
                    value = self.getData(dependency)
                    if value is None: isNull = True; break
                    vars[dependency] = value
                if isNull:
                    continue
                expr = rule.getExpr()
                # Now evaluate the rule to see if its violated
                try:
                    vars[k] = v # test it with new value from web
                    isOkay = expr.evaluate(vars)
                except:
                    print_exc()
                    isOkay = False
                if not isOkay:
                    hasNoProblems = False
                    ie = old_errors.get(k, None)
                    if ie is None:
                        try:
                            desc = widgets_clone[k].getDescription()
                        except KeyError:
                            desc = ""
                        ie = ErrorItem(pageName, v, desc)
                        old_errors[k] = ie
                    msg = rule.getValue()
                    if msg: ie.addMessage(msg)
                    else:   ie.addMessage("There is a problem with '%s'." % k)
            if hasNoProblems: # No new errors -> set it
                if dbPage.setData(k, v):
                    dirty = True
                # TODO: Solve transitive problems!
                # e.g. A was set. B contradicts A. That prevents B from
                # being set. -> B is marked was wrong input.
                # Now A is set to a value that does not contradicts B
                # any more. Therefore B is not an error input any more.
                # -> B can be applied. This scheme has to be applied 
                # recursively along the dependencies.
                old_errors.pop(k, None)

        if dirty: # flush the db page
            self.dirty = True
            self.update_inconsistencies()

        # Before continuing write a swap file, but do not overwrite the backup
        try:
            self.store(self.swapfile, backup=False)
        except:
            pass
        return old_errors


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
